@@ -29,11 +29,13 @@ import java.util.ArrayList;
 import java.util.Collections;
 import java.util.HashSet;
 import java.util.List;
+import java.util.Set;
 
 import javax.xml.parsers.ParserConfigurationException;
 
 import org.xml.sax.SAXException;
 
+import beast.core.Input.Validate;
 import beast.core.util.CompoundDistribution;
 import beast.core.util.Evaluator;
 import beast.core.util.Log;
@@ -42,12 +44,21 @@ import beast.util.Randomizer;
 @Description("MCMC chain. This is the main element that controls which posterior " +
         "to calculate, how long to run the chain and all other properties, " +
         "which operators to apply on the state space and where to log results.")
-@Citation(value=
-        "Bouckaert RR, Heled J, Kuehnert D, Vaughan TG, Wu C-H, Xie D, Suchard MA,\n" +
-                "  Rambaut A, Drummond AJ (2014) BEAST 2: A software platform for Bayesian\n" +
-                "  evolutionary analysis. PLoS Computational Biology 10(4): e1003537"
-        , year = 2014, firstAuthorSurname = "bouckaert",
-        DOI="10.1371/journal.pcbi.1003537")
+//@Citation(value=
+//        "Bouckaert RR, Heled J, Kuehnert D, Vaughan TG, Wu C-H, Xie D, Suchard MA,\n" +
+//                "  Rambaut A, Drummond AJ (2014) BEAST 2: A software platform for Bayesian\n" +
+//                "  evolutionary analysis. PLoS Computational Biology 10(4): e1003537"
+//        , year = 2014, firstAuthorSurname = "bouckaert",
+//        DOI="10.1371/journal.pcbi.1003537")
+@Citation(value="Bouckaert, Remco, Timothy G. Vaughan, Joëlle Barido-Sottani, Sebastián Duchêne, Mathieu Fourment, \n"
+		+ "Alexandra Gavryushkina, Joseph Heled, Graham Jones, Denise Kühnert, Nicola De Maio, Michael Matschiner, \n"
+        + "Fábio K. Mendes, Nicola F. Müller, Huw A. Ogilvie, Louis du Plessis, Alex Popinga, Andrew Rambaut, \n"
+        + "David Rasmussen, Igor Siveroni, Marc A. Suchard, Chieh-Hsi Wu, Dong Xie, Chi Zhang, Tanja Stadler, \n"
+        + "Alexei J. Drummond \n"
+		+ "  BEAST 2.5: An advanced software platform for Bayesian evolutionary analysis. \n"
+		+ "  PLoS computational biology 15, no. 4 (2019): e1006650.", 
+        year = 2019, firstAuthorSurname = "bouckaert",
+		DOI="10.1371/journal.pcbi.1006650")
 public class MCMC extends Runnable {
 
     final public Input<Long> chainLengthInput =
@@ -88,7 +99,7 @@ public class MCMC extends Runnable {
     final public Input<Boolean> sampleFromPriorInput = new Input<>("sampleFromPrior", "whether to ignore the likelihood when sampling (default false). " +
             "The distribution with id 'likelihood' in the posterior input will be ignored when this flag is set.", false);
 
-    final public Input<OperatorSchedule> operatorScheduleInput = new Input<>("operatorschedule", "specify operator selection and optimisation schedule", new OperatorSchedule());
+    final public Input<OperatorSchedule> operatorScheduleInput = new Input<>("operatorschedule", "specify operator selection and optimisation schedule", new OperatorSchedule(), Validate.REQUIRED);
 
     /**
      * Alternative representation of operatorsInput that allows random selection
@@ -113,7 +124,7 @@ public class MCMC extends Runnable {
      */
     final protected int NR_OF_DEBUG_SAMPLES = 2000;
 
-    /**
+    /** 
      * Interval for storing state to disk, if negative the state will not be stored periodically *
      * Mirrors m_storeEvery input, or if this input is negative, the State.m_storeEvery input
      */
@@ -346,6 +357,9 @@ public class MCMC extends Runnable {
         }
 
         // initialises log so that log file headers are written, etc.
+        if (restoreFromFile) {
+        	makeSureLogFilesAreSameLength();
+        }
         for (final Logger log : loggers) {
             log.init();
         }
@@ -368,7 +382,32 @@ public class MCMC extends Runnable {
     } // run;
 
 
-    /**
+    protected void makeSureLogFilesAreSameLength() throws IOException {
+    	// make sure log files all end in the same state
+    	long min = -1;
+    	long [] lengths = new long[loggers.size()];
+    	for (int i = 0; i < loggers.size(); i++) {
+    		Logger logger = loggers.get(i);
+    		if (!logger.isLoggingToStdout()) {
+    			long offset = logger.getLogOffset();
+    			lengths[i] = offset;
+    			min = min == -1 ? offset : Math.min(min, offset); 
+    		}
+    	}
+    	if (min == 0) {
+    		// at least one log file is empty
+    		return;
+    	}
+    	for (int i = 0; i < loggers.size(); i++) {
+    		Logger logger = loggers.get(i);
+    		if (!logger.isLoggingToStdout() && min != lengths[i]) {
+    			logger.setLogOffset(min);
+    		}
+    	}
+	}
+
+
+	/**
      * main MCMC loop 
      * @throws IOException *
      */
@@ -411,7 +450,7 @@ public class MCMC extends Runnable {
                     }
                 } else {
                     if (isTooDifferent(logLikelihood, originalLogP)) {
-                        // halt due to incorrect posterior during intial debug period
+                        // halt due to incorrect posterior during initial debug period
                         state.storeToFile(sampleNr);
                         operatorSchedule.storeToFile();
                         System.exit(1);
@@ -484,6 +523,15 @@ public class MCMC extends Runnable {
                 }
             };
         }
+
+        if (debugFlag) {
+            // Store a checksum of each calculation node to ensure that the initial
+            // state is restored when a step is rejected (see validateReject).
+            for (CalculationNode node : listAllCalcNodes()) {
+                node.storeChecksum();
+            }
+        }
+
         final double logHastingsRatio = operator.proposal(evaluator);
 
         if (logHastingsRatio != Double.NEGATIVE_INFINITY) {
@@ -515,6 +563,9 @@ public class MCMC extends Runnable {
                 state.restore();
                 state.restoreCalculationNodes();
                 if (printDebugInfo) System.err.print(" reject");
+                if (debugFlag) {
+                    validateReject(operator);
+                }
             }
             state.setEverythingDirty(false);
         } else {
@@ -528,11 +579,62 @@ public class MCMC extends Runnable {
                 state.restoreCalculationNodes();
             }
             if (printDebugInfo) System.err.print(" direct reject");
+            if (debugFlag) {
+                validateReject(operator);
+            }
         }
         log(sampleNr);
         return operator;
     }
 
+    /**
+     * Test whether state nodes and fat calculation nodes have been correctly
+     * restored (according to their checksums) after the operator was rejected.
+     * Caching an incorrect state could lead to undetected errors, here we
+     * explicitly throw an exception if something went wrong.
+     *
+     * @param operator: the operator applied in the current MCMC step.
+     */
+    private void validateReject(Operator operator) throws RuntimeException {
+        List<CalculationNode> affectedNodes = new ArrayList<>();
+        // Relevant nodes: all calculation nodes that have changed in this step
+        affectedNodes.addAll(state.getCurrentCalculationNodes());
+        // ...and all state nodes that are affected by the operator.
+        affectedNodes.addAll(operator.listStateNodes());
+
+        for (CalculationNode node : affectedNodes) {
+            if (!node.matchesOldChecksum()) {
+                throw new RuntimeException("Node " + node.getID() + " was incorrectly restored after rejecting "
+                        + "operator " + operator.getID() + " (according to calculatioNode checksum).");
+            }
+        }
+    }
+
+    /**
+     * Recursively iterate over all inputs of the BEASTInterface o and
+     * add the calculationNodes to allNodes.
+     * @param o the BEASTInterface object at which the recursive search is started
+     * @param allNodes the list where the calculationNodes are collected.
+     **/
+    protected void collectCalcNodes(BEASTInterface o, Set<CalculationNode> allNodes) {
+        if (o instanceof CalculationNode) {
+            allNodes.add((CalculationNode) o);
+        }
+        for (BEASTInterface o2 : o.listActiveBEASTObjects()) {
+            collectCalcNodes(o2, allNodes);
+        }
+    }
+
+    /**
+     * Collect all calculation nodes used in the posterior in a list.
+     * @return The list of all calculation nodes.
+     */
+    protected Set<CalculationNode> listAllCalcNodes() {
+        Set<CalculationNode> allNodes = new HashSet<>();
+        collectCalcNodes(posterior, allNodes);
+        return allNodes;
+    }
+
     private boolean isTooDifferent(double logLikelihood, double originalLogP) {
     	//return Math.abs((logLikelihood - originalLogP)/originalLogP) > 1e-6;
     	return Math.abs(logLikelihood - originalLogP) > 1e-6;
